==> BlueWhale3-3.28.11-cp36-cp36m-macosx_10_9_x86_64.whl: Orange/widgets/unsupervised/owdistances.py <==
from AnyQt.QtCore import Qt
from scipy.sparse import issparse
import bottleneck as bn

import Orange.data
import Orange.misc
from Orange import distance
from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.utils.concurrent import TaskState, ConcurrentWidgetMixin
from Orange.widgets.utils.sql import check_sql_input
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.utils.state_summary import format_summary_details
from Orange.widgets.widget import OWWidget, Msg, Input, Output

from Orange.i18n_config import *


def __(key):
    return i18n.t("widget.unsupervised.unsupervised.owdistances." + key)


METRICS = [
    (i18n.t("common.algorithm.euclidean"), distance.Euclidean),
    (i18n.t("common.algorithm.manhattan"), distance.Manhattan),
    (i18n.t("common.algorithm.cosine"), distance.Cosine),
    (i18n.t("common.algorithm.jaccard"), distance.Jaccard),
    (i18n.t("common.algorithm.spearman"), distance.SpearmanR),
    (i18n.t("common.algorithm.absolute_spearman"), distance.SpearmanRAbsolute),
    (i18n.t("common.algorithm.pearson"), distance.PearsonR),
    (i18n.t("common.algorithm.absolute_pearson"), distance.PearsonRAbsolute),
    (i18n.t("common.algorithm.hamming"), distance.Hamming),
    (i18n.t("common.algorithm.mahalanobis"), distance.Mahalanobis),
    (i18n.t("common.algorithm.bhattacharyya"), distance.Bhattacharyya)
]


class InterruptException(Exception):
    pass


class DistanceRunner:
    @staticmethod
    def run(data: Orange.data.Table, metric: distance, normalized_dist: bool,
            axis: int, state: TaskState) -> Orange.misc.DistMatrix:
        if data is None:
            return None

        def callback(i: float) -> bool:
            state.set_progress_value(i)
            if state.is_interruption_requested():
                raise InterruptException

        state.set_status(__("set_status_calculating"))
        kwargs = {"axis": 1 - axis, "impute": True, "callback": callback}
        if metric.supports_normalization and normalized_dist:
            kwargs["normalize"] = True
        return metric(data, **kwargs)


class OWDistances(OWWidget, ConcurrentWidgetMixin):
    name = __("name")
    description = __("desc")
    icon = "icons/Distance.svg"
    keywords = []

    class Inputs:
        data = Input('Data', Orange.data.Table, label=i18n.t("widget.unsupervised.unsupervised.common.data"))

    class Outputs:
        distances = Output('Distances', Orange.misc.DistMatrix, dynamic=False,
                           label=i18n.t("common.general.distance"))

    settings_version = 3

    axis = Setting(0)  # type: int
    metric_idx = Setting(0)  # type: int

    #: Use normalized distances if the metric supports it.
    #: The default is `True`, expect when restoring from old pre v2 settings
    #: (see `migrate_settings`).
    normalized_dist = Setting(True)  # type: bool
    autocommit = Setting(True)  # type: bool

    want_main_area = False
    resizing_enabled = False
    buttons_area_orientation = Qt.Vertical

    class Error(OWWidget.Error):
        no_continuous_features = Msg(__("msg.no_numeric"))
        no_binary_features = Msg(__("msg.no_binary"))
        dense_metric_sparse_data = Msg(__("msg.dense_data"))
        distances_memory_error = Msg(i18n.t("common.software.not_enough_memory"))
        distances_value_error = Msg(__("msg.problem"))
        data_too_large_for_mahalanobis = Msg(__("msg.upto"))

    class Warning(OWWidget.Warning):
        ignoring_discrete = Msg(__("msg.ignoring_categorical"))
        ignoring_nonbinary = Msg(__("msg.ignoring_non-binary"))
        imputing_data = Msg(__("msg.miss_value"))

    def __init__(self):
        OWWidget.__init__(self)
        ConcurrentWidgetMixin.__init__(self)

        self.data = None

        self._set_input_summary(None)
        self._set_output_summary(None)

        gui.radioButtons(
            self.controlArea, self, "axis", [i18n.t("common.general.row"), i18n.t("common.general.column")],
            box=__("box_distances_between"), callback=self._invalidate
        )
        box = gui.widgetBox(self.controlArea, __("box_distance_metric"))
        self.metrics_combo = gui.comboBox(
            box, self, "metric_idx",
            items=[m[0] for m in METRICS],
            callback=self._metric_changed
        )
        self.normalization_check = gui.checkBox(
            box, self, "normalized_dist", __("row.normalized"),
            callback=self._invalidate,
            tooltip=(__("tooltip_influence_domain")),
            stateWhenDisabled=False
        )
        _, metric = METRICS[self.metric_idx]
        self.normalization_check.setEnabled(metric.supports_normalization)

        gui.auto_apply(self.controlArea, self, "autocommit")

    @Inputs.data
    @check_sql_input
    def set_data(self, data):
        self.cancel()
        self.data = data
        self._set_input_summary(data)
        self.refresh_metrics()
        self.unconditional_commit()

    def refresh_metrics(self):
        sparse = self.data is not None and issparse(self.data.X)
        for i, metric in enumerate(METRICS):
            item = self.metrics_combo.model().item(i)
            item.setEnabled(not sparse or metric[1].supports_sparse)

    def commit(self):
        # pylint: disable=invalid-sequence-index
        metric = METRICS[self.metric_idx][1]
        self.compute_distances(metric, self.data)

    def compute_distances(self, metric, data):
        def _check_sparse():
            # pylint: disable=invalid-sequence-index
            if issparse(data.X) and not metric.supports_sparse:
                self.Error.dense_metric_sparse_data(METRICS[self.metric_idx][0])
                return False
            return True

        def _fix_discrete():
            nonlocal data
            if data.domain.has_discrete_attributes() \
                    and metric is not distance.Jaccard \
                    and (issparse(data.X) and getattr(metric, "fallback", None)
                         or not metric.supports_discrete
                         or self.axis == 1):
                if not data.domain.has_continuous_attributes():
                    self.Error.no_continuous_features()
                    return False
                self.Warning.ignoring_discrete()
                data = distance.remove_discrete_features(data)
            return True

        def _fix_nonbinary():
            nonlocal data
            if metric is distance.Jaccard and not issparse(data.X):
                nbinary = sum(a.is_discrete and len(a.values) == 2
                              for a in data.domain.attributes)
                if not nbinary:
                    self.Error.no_binary_features()
                    return False
                elif nbinary < len(data.domain.attributes):
                    self.Warning.ignoring_nonbinary()
                    data = distance.remove_nonbinary_features(data)
            return True

        def _fix_missing():
            nonlocal data
            if not metric.supports_missing and bn.anynan(data.X):
                self.Warning.imputing_data()
                data = distance.impute(data)
            return True

        def _check_tractability():
            if metric is distance.Mahalanobis:
                if self.axis == 1:
                    # when computing distances by columns, we want < 100 rows
                    if len(data) > 1000:
                        self.Error.data_too_large_for_mahalanobis("rows")
                        return False
                else:
                    if len(data.domain.attributes) > 1000:
                        self.Error.data_too_large_for_mahalanobis("columns")
                        return False
            return True

        self.clear_messages()
        if data is not None:
            for check in (_check_sparse, _check_tractability,
                          _fix_discrete, _fix_missing, _fix_nonbinary):
                if not check():
                    data = None
                    break

        self.start(DistanceRunner.run, data, metric,
                   self.normalized_dist, self.axis)

    def on_partial_result(self, _):
        pass

    def on_done(self, result: Orange.misc.DistMatrix):
        assert isinstance(result, Orange.misc.DistMatrix) or result is None
        self._set_output_summary(result)
        self.Outputs.distances.send(result)

    def on_exception(self, ex):
        if isinstance(ex, ValueError):
            self.Error.distances_value_error(ex)
        elif isinstance(ex, MemoryError):
            self.Error.distances_memory_error()
        elif isinstance(ex, InterruptException):
            pass
        else:
            raise ex

    def _set_input_summary(self, data):
        summary = len(data) if data else self.info.NoInput
        details = format_summary_details(data) if data else ""
        self.info.set_input_summary(summary, details)

    def _set_output_summary(self, output):
        if output is None:
            self.info.set_output_summary(self.info.NoOutput)
        else:
            summary = f"{output.shape[0]}×{output.shape[1]}"
            self.info.set_output_summary(summary, summary)

    def onDeleteWidget(self):
        self.shutdown()
        super().onDeleteWidget()

    def _invalidate(self):
        self.commit()

    def _metric_changed(self):
        metric = METRICS[self.metric_idx][1]
        self.normalization_check.setEnabled(metric.supports_normalization)
        self._invalidate()

    def send_report(self):
        # pylint: disable=invalid-sequence-index
        self.report_items((
            (__("row.distances_between"), [i18n.t("common.general.row"), __("row.columns")][self.axis]),
            (i18n.t("common.general.metric"), METRICS[self.metric_idx][0])
        ))

    @classmethod
    def migrate_settings(cls, settings, version):
        if version is None or version < 2 and "normalized_dist" not in settings:
            # normalize_dist is set to False when restoring settings from
            # an older version to preserve old semantics.
            settings["normalized_dist"] = False
        if version is None or version < 3:
            # Mahalanobis was moved from idx = 2 to idx = 9
            metric_idx = settings["metric_idx"]
            if metric_idx == 2:
                settings["metric_idx"] = 9
            elif 2 < metric_idx <= 9:
                settings["metric_idx"] -= 1


if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWDistances).run(Orange.data.Table("iris"))

==> BlueWhale3-3.29.0-cp36-cp36m-macosx_10_9_x86_64.whl: Orange/widgets/unsupervised/owdistances.py <==
from scipy.sparse import issparse
import bottleneck as bn

from AnyQt.QtCore import Qt

import Orange.data
import Orange.misc
from Orange import distance
from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.utils.concurrent import TaskState, ConcurrentWidgetMixin
from Orange.widgets.utils.sql import check_sql_input
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.utils.state_summary import format_summary_details
from Orange.widgets.widget import OWWidget, Msg, Input, Output

from Orange.i18n_config import *


def __(key):
    return i18n.t("widget.unsupervised.unsupervised.owdistances." + key)


METRICS = [
    (i18n.t("common.algorithm.euclidean"), distance.Euclidean),
    (i18n.t("common.algorithm.manhattan"), distance.Manhattan),
    (i18n.t("common.algorithm.cosine"), distance.Cosine),
    (i18n.t("common.algorithm.jaccard"), distance.Jaccard),
    (i18n.t("common.algorithm.spearman"), distance.SpearmanR),
    (i18n.t("common.algorithm.absolute_spearman"), distance.SpearmanRAbsolute),
    (i18n.t("common.algorithm.pearson"), distance.PearsonR),
    (i18n.t("common.algorithm.absolute_pearson"), distance.PearsonRAbsolute),
    (i18n.t("common.algorithm.hamming"), distance.Hamming),
    (i18n.t("common.algorithm.mahalanobis"), distance.Mahalanobis),
    (i18n.t("common.algorithm.bhattacharyya"), distance.Bhattacharyya)
]


class InterruptException(Exception):
    pass


class DistanceRunner:
    @staticmethod
    def run(data: Orange.data.Table, metric: distance, normalized_dist: bool,
            axis: int, state: TaskState) -> Orange.misc.DistMatrix:
        if data is None:
            return None

        def callback(i: float) -> bool:
            state.set_progress_value(i)
            if state.is_interruption_requested():
                raise InterruptException

        state.set_status(__("set_status_calculating"))
        kwargs = {"axis": 1 - axis, "impute": True, "callback": callback}
        if metric.supports_normalization and normalized_dist:
            kwargs["normalize"] = True
        return metric(data, **kwargs)


class OWDistances(OWWidget, ConcurrentWidgetMixin):
    name = __("name")
    description = __("desc")
    icon = "icons/Distance.svg"
    keywords = []

    class Inputs:
        data = Input('Data', Orange.data.Table, label=i18n.t("widget.unsupervised.unsupervised.common.data"))

    class Outputs:
        distances = Output('Distances', Orange.misc.DistMatrix, dynamic=False,
                           label=i18n.t("common.general.distance"))

    settings_version = 3

    axis = Setting(0)  # type: int
    metric_idx = Setting(0)  # type: int

    #: Use normalized distances if the metric supports it.
    #: The default is `True`, expect when restoring from old pre v2 settings
    #: (see `migrate_settings`).
    normalized_dist = Setting(True)  # type: bool
    autocommit = Setting(True)  # type: bool

    want_main_area = False
    resizing_enabled = False

    class Error(OWWidget.Error):
        no_continuous_features = Msg(__("msg.no_numeric"))
        no_binary_features = Msg(__("msg.no_binary"))
        dense_metric_sparse_data = Msg(__("msg.dense_data"))
        distances_memory_error = Msg(i18n.t("common.software.not_enough_memory"))
        distances_value_error = Msg(__("msg.problem"))
        data_too_large_for_mahalanobis = Msg(__("msg.upto"))

    class Warning(OWWidget.Warning):
        ignoring_discrete = Msg(__("msg.ignoring_categorical"))
        ignoring_nonbinary = Msg(__("msg.ignoring_non-binary"))
        imputing_data = Msg(__("msg.miss_value"))

    def __init__(self):
        OWWidget.__init__(self)
        ConcurrentWidgetMixin.__init__(self)

        self.data = None

        self._set_input_summary(None)
        self._set_output_summary(None)

        gui.radioButtons(
            self.controlArea, self, "axis", [i18n.t("common.general.row"), i18n.t("common.general.column")],
            box=__("box_distances_between"), callback=self._invalidate
        )
        box = gui.widgetBox(self.controlArea, __("box_distance_metric"))
        self.metrics_combo = gui.comboBox(
            box, self, "metric_idx",
            items=[m[0] for m in METRICS],
            callback=self._metric_changed
        )
        self.normalization_check = gui.checkBox(
            box, self, "normalized_dist", __("row.normalized"),
            callback=self._invalidate,
            tooltip=(__("tooltip_influence_domain")),
            stateWhenDisabled=False, attribute=Qt.WA_LayoutUsesWidgetRect
        )
        _, metric = METRICS[self.metric_idx]
        self.normalization_check.setEnabled(metric.supports_normalization)

        gui.auto_apply(self.buttonsArea, self, "autocommit")

    @Inputs.data
    @check_sql_input
    def set_data(self, data):
        self.cancel()
        self.data = data
        self._set_input_summary(data)
        self.refresh_metrics()
        self.unconditional_commit()

    def refresh_metrics(self):
        sparse = self.data is not None and issparse(self.data.X)
        for i, metric in enumerate(METRICS):
            item = self.metrics_combo.model().item(i)
            item.setEnabled(not sparse or metric[1].supports_sparse)

    def commit(self):
        # pylint: disable=invalid-sequence-index
        metric = METRICS[self.metric_idx][1]
        self.compute_distances(metric, self.data)

    def compute_distances(self, metric, data):
        def _check_sparse():
            # pylint: disable=invalid-sequence-index
            if issparse(data.X) and not metric.supports_sparse:
                self.Error.dense_metric_sparse_data(METRICS[self.metric_idx][0])
                return False
            return True

        def _fix_discrete():
            nonlocal data
            if data.domain.has_discrete_attributes() \
                    and metric is not distance.Jaccard \
                    and (issparse(data.X) and getattr(metric, "fallback", None)
                         or not metric.supports_discrete
                         or self.axis == 1):
                if not data.domain.has_continuous_attributes():
                    self.Error.no_continuous_features()
                    return False
                self.Warning.ignoring_discrete()
                data = distance.remove_discrete_features(data)
            return True

        def _fix_nonbinary():
            nonlocal data
            if metric is distance.Jaccard and not issparse(data.X):
                nbinary = sum(a.is_discrete and len(a.values) == 2
                              for a in data.domain.attributes)
                if not nbinary:
                    self.Error.no_binary_features()
                    return False
                elif nbinary < len(data.domain.attributes):
                    self.Warning.ignoring_nonbinary()
                    data = distance.remove_nonbinary_features(data)
            return True

        def _fix_missing():
            nonlocal data
            if not metric.supports_missing and bn.anynan(data.X):
                self.Warning.imputing_data()
                data = distance.impute(data)
            return True

        def _check_tractability():
            if metric is distance.Mahalanobis:
                if self.axis == 1:
                    # when computing distances by columns, we want < 100 rows
                    if len(data) > 1000:
                        self.Error.data_too_large_for_mahalanobis("rows")
                        return False
                else:
                    if len(data.domain.attributes) > 1000:
                        self.Error.data_too_large_for_mahalanobis("columns")
                        return False
            return True

        self.clear_messages()
        if data is not None:
            for check in (_check_sparse, _check_tractability,
                          _fix_discrete, _fix_missing, _fix_nonbinary):
                if not check():
                    data = None
                    break

        self.start(DistanceRunner.run, data, metric,
                   self.normalized_dist, self.axis)

    def on_partial_result(self, _):
        pass

    def on_done(self, result: Orange.misc.DistMatrix):
        assert isinstance(result, Orange.misc.DistMatrix) or result is None
        self._set_output_summary(result)
        self.Outputs.distances.send(result)

    def on_exception(self, ex):
        if isinstance(ex, ValueError):
            self.Error.distances_value_error(ex)
        elif isinstance(ex, MemoryError):
            self.Error.distances_memory_error()
        elif isinstance(ex, InterruptException):
            pass
        else:
            raise ex

    def _set_input_summary(self, data):
        summary = len(data) if data else self.info.NoInput
        details = format_summary_details(data) if data else ""
        self.info.set_input_summary(summary, details)

    def _set_output_summary(self, output):
        if output is None:
            self.info.set_output_summary(self.info.NoOutput)
        else:
            summary = f"{output.shape[0]}×{output.shape[1]}"
            self.info.set_output_summary(summary, summary)

    def onDeleteWidget(self):
        self.shutdown()
        super().onDeleteWidget()

    def _invalidate(self):
        self.commit()

    def _metric_changed(self):
        metric = METRICS[self.metric_idx][1]
        self.normalization_check.setEnabled(metric.supports_normalization)
        self._invalidate()

    def send_report(self):
        # pylint: disable=invalid-sequence-index
        self.report_items((
            (__("row.distances_between"), [i18n.t("common.general.row"), __("row.columns")][self.axis]),
            (i18n.t("common.general.metric"), METRICS[self.metric_idx][0])
        ))

    @classmethod
    def migrate_settings(cls, settings, version):
        if version is None or version < 2 and "normalized_dist" not in settings:
            # normalize_dist is set to False when restoring settings from
            # an older version to preserve old semantics.
            settings["normalized_dist"] = False
        if version is None or version < 3:
            # Mahalanobis was moved from idx = 2 to idx = 9
            metric_idx = settings["metric_idx"]
            if metric_idx == 2:
                settings["metric_idx"] = 9
            elif 2 < metric_idx <= 9:
                settings["metric_idx"] -= 1


if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWDistances).run(Orange.data.Table("iris"))
